.feature force_range

;
; iNES header
;

.segment "HEADER"

INES_MAPPER = 0 ; 0 = NROM
INES_MIRROR = 1 ; 0 = horizontal mirroring, 1 = vertical mirroring
INES_SRAM   = 0 ; 1 = battery backed SRAM at $6000-7FFF

.byte 'N', 'E', 'S', $1A ; ID
.byte $02 ; 16k PRG bank count
.byte $01 ; 8k CHR bank count
.byte INES_MIRROR | (INES_SRAM << 1) | ((INES_MAPPER & $f) << 4)
.byte (INES_MAPPER & %11110000)
.byte $0, $0, $0, $0, $0, $0, $0, $0 ; padding

;
; CHR ROM
;

.segment "TILES"
.incbin "art/fp_bg.chr"
.incbin "art/fp_spr.chr"

;
; vectors placed at top 6 bytes of memory area
;

.segment "VECTORS"
.word nmi
.word reset
.word irq

;
; reset routine
;

.segment "CODE"
reset:
	sei       ; mask interrupts
	lda #0
	sta $2000 ; disable NMI
	sta $2001 ; disable rendering
	sta $4015 ; disable APU sound
	sta $4010 ; disable DMC IRQ
	lda #$40
	sta $4017 ; disable APU IRQ
	cld       ; disable decimal mode
	ldx #$FF
	txs       ; initialize stack
	; wait for first vblank
	bit $2002
	:
		bit $2002
		bpl :-
	; save random number seed
	lda prng_seed+0
	pha
	lda prng_seed+1
	pha
	; clear all RAM to 0 (except $100 stack area)
	lda #0
	ldx #0
	:
		sta $0000, X
		sta $0200, X
		sta $0300, X
		sta $0400, X
		sta $0500, X
		sta $0600, X
		sta $0700, X
		inx
		bne :-
	; restore random number seed
	pla
	sta prng_seed+1
	pla
	sta prng_seed+0
	; place all sprites offscreen at Y=255
	lda #255
	ldx #0
	:
		sta oam, X
		inx
		inx
		inx
		inx
		bne :-
	; wait for second vblank
	:
		bit $2002
		bpl :-
	; NES is initialized, ready to begin!
	; enable the NMI for graphical updates, and jump to our main program
	lda #%10001000
	sta $2000
	jmp main

;
; nmi routine
;

.segment "ZEROPAGE"
nmi_lock:       .res 1 ; prevents NMI re-entry
nmi_count:      .res 1 ; is incremented every NMI
nmi_ready:      .res 1 ; set to 1 to push a PPU frame update, 2 to turn rendering off next NMI
nmt_update_len: .res 1 ; number of bytes in nmt_update buffer
scroll_x:       .res 1 ; x scroll position
scroll_y:       .res 1 ; y scroll position
scroll_nmt:     .res 1 ; nametable select (0-3 = $2000,$2400,$2800,$2C00)
temp:           .res 1 ; temporary variable

.segment "RAM"
nmt_update: .res 256 ; nametable update entry buffer for PPU update
palette:    .res 32  ; palette buffer for PPU update

.segment "OAM"
oam: .res 256        ; sprite OAM data to be uploaded by DMA

.segment "CODE"
nmi:
	; save registers
	pha
	txa
	pha
	tya
	pha
	; prevent NMI re-entry
	lda nmi_lock
	beq :+
		jmp @nmi_end
	:
	lda #1
	sta nmi_lock
	; increment frame counter
	inc nmi_count
	;
	lda nmi_ready
	bne :+ ; nmi_ready == 0 not ready to update PPU
		jmp @ppu_update_end
	:
	cmp #2 ; nmi_ready == 2 turns rendering off
	bne :+
		lda #%00000000
		sta $2001
		ldx #0
		stx nmi_ready
		jmp @ppu_update_end
	:
	; sprite OAM DMA
	ldx #0
	stx $2003
	lda #>oam
	sta $4014
	; palettes
	lda #%10001000
	sta $2000 ; set horizontal nametable increment
	lda $2002
	lda #$3F
	sta $2006
	stx $2006 ; set PPU address to $3F00
	ldx #0
	:
		lda palette, X
		sta $2007
		inx
		cpx #32
		bcc :-
	; nametable update
	ldx #0
	cpx nmt_update_len
	bcs @scroll
	@nmt_update_loop:
		lda nmt_update, X
		sta $2006
		inx
		lda nmt_update, X
		sta $2006
		inx
		lda nmt_update, X
		sta $2007
		inx
		cpx nmt_update_len
		bcc @nmt_update_loop
	lda #0
	sta nmt_update_len
@scroll:
	lda scroll_nmt
	and #%00000011 ; keep only lowest 2 bits to prevent error
	ora #%10001000
	sta $2000
	lda scroll_x
	sta $2005
	lda scroll_y
	sta $2005
	; enable rendering
	lda #%00011110
	sta $2001
	; flag PPU update complete
	ldx #0
	stx nmi_ready
@ppu_update_end:
	; if this engine had music/sound, this would be a good place to play it
	; unlock re-entry flag
	lda #0
	sta nmi_lock
@nmi_end:
	; restore registers and return
	pla
	tay
	pla
	tax
	pla
	rti

;
; irq
;

.segment "CODE"
irq:
	rti

;
; drawing utilities
;

.segment "CODE"

; ppu_update: waits until next NMI, turns rendering on (if not already), uploads OAM, palette, and nametable update to PPU
ppu_update:
	;lda #%00011111 ; greyscale
	;sta $2001 ; profile performance with greyscale change
	lda #1
	sta nmi_ready
	:
		lda nmi_ready
		bne :-
	rts

; ppu_skip: waits until next NMI, does not update PPU
ppu_skip:
	lda nmi_count
	:
		cmp nmi_count
		beq :-
	rts

; ppu_off: waits until next NMI, turns rendering off (now safe to write PPU directly via $2007)
ppu_off:
	lda #2
	sta nmi_ready
	:
		lda nmi_ready
		bne :-
	rts

; ppu_address_tile: use with rendering off, sets memory address to tile at X/Y, ready for a $2007 write
;   Y =  0- 31 nametable $2000
;   Y = 32- 63 nametable $2400
;   Y = 64- 95 nametable $2800
;   Y = 96-127 nametable $2C00
ppu_address_tile:
	lda $2002 ; reset latch
	tya
	lsr
	lsr
	lsr
	ora #$20 ; high bits of Y + $20
	sta $2006
	tya
	asl
	asl
	asl
	asl
	asl
	sta temp
	txa
	ora temp
	sta $2006 ; low bits of Y + X
	rts

; ppu_update_tile: can be used with rendering on, sets the tile at X/Y to tile A next time you call ppu_update
ppu_update_tile:
	pha ; temporarily store A on stack
	txa
	pha ; temporarily store X on stack
	ldx nmt_update_len
	tya
	lsr
	lsr
	lsr
	ora #$20 ; high bits of Y + $20
	sta nmt_update, X
	inx
	tya
	asl
	asl
	asl
	asl
	asl
	sta temp
	pla ; recover X value (but put in A)
	ora temp
	sta nmt_update, X
	inx
	pla ; recover A value (tile)
	sta nmt_update, X
	inx
	stx nmt_update_len
	rts

; ppu_update_byte: like ppu_update_tile, but X/Y makes the high/low bytes of the PPU address to write
;    this may be useful for updating attribute tiles
ppu_update_byte:
	pha ; temporarily store A on stack
	tya
	pha ; temporarily store Y on stack
	ldy nmt_update_len
	txa
	sta nmt_update, Y
	iny
	pla ; recover Y value (but put in Y)
	sta nmt_update, Y
	iny
	pla ; recover A value (byte)
	sta nmt_update, Y
	iny
	sty nmt_update_len
	rts

;
; gamepad
;

PAD_A      = $01
PAD_B      = $02
PAD_SELECT = $04
PAD_START  = $08
PAD_U      = $10
PAD_D      = $20
PAD_L      = $40
PAD_R      = $80

.segment "ZEROPAGE"
gamepad: .res 1

.segment "CODE"
; gamepad_poll: this reads the gamepad state into the variable labelled "gamepad"
;   This only reads the first gamepad, and also if DPCM samples are played they can
;   conflict with gamepad reading, which may give incorrect results.
gamepad_poll:
	; strobe the gamepad to latch current button state
	lda #1
	sta $4016
	lda #0
	sta $4016
	; read 8 bytes from the interface at $4016
	ldx #8
	:
		pha
		lda $4016
		; combine low two bits and store in carry bit
		and #%00000011
		cmp #%00000001
		pla
		; rotate carry into gamepad variable
		ror
		dex
		bne :-
	sta gamepad
	rts

;
; main
;

.segment "RODATA"
fp_palette:
.incbin "art/fp.pal"
.incbin "art/fp.pal"

fp_title:
.incbin "art/fp_title.nam"

particle00: .byte    0,  -2,$02,$00, 128 ; coloured dot, palette 0,1,2,3
particle01: .byte    0,  -2,$03,$00, 128 ; white dot
particle02: .byte    0,  -4,$04,$00, 128 ; white cross
particle03: .byte    0,  -4,$05,$00, 128 ; white open cross
particle04: .byte    0,  -8,$06,$02, 128 ; star, palette 2,3
particle05: .byte    0, -14,$07,$02      ; big star, palette 2,3
            .byte    5, -14,$07,$42
            .byte    0,  -6,$17,$02
            .byte    5,  -6,$17,$42, 128
particle06: .byte    0,  -8,$10,$00, 128 ; white shiny 0
particle07: .byte    0,  -8,$11,$00, 128 ; white shiny 1
particle08: .byte    0, -26,$20,$00      ; bubble huge, palette 0,1,2,3
            .byte    8, -26,$24,$00
            .byte    9, -26,$24,$40
            .byte   17, -26,$21,$40
            .byte    0, -18,$22,$00
            .byte   17, -18,$23,$40
            .byte    0, -17,$23,$80
            .byte   17, -17,$23,$C0
            .byte    0,  -9,$21,$80
            .byte    8,  -9,$24,$80
            .byte    9,  -9,$24,$C0
            .byte   17,  -9,$25,$00, 128
particle09: .byte    0, -24,$30,$00      ; bubble large, palette 0,1,2,3
            .byte    8, -24,$32,$00
            .byte   15, -24,$31,$40
            .byte    0, -16,$33,$00
            .byte   15, -16,$33,$40
            .byte    0,  -9,$31,$80
            .byte    8,  -9,$32,$80
            .byte   15,  -9,$31,$C0, 128
particle0A: .byte    0, -16,$40,$00      ; bubble medium, palette 0,1,2,3
            .byte    7, -16,$41,$40
            .byte    0,  -9,$41,$80
            .byte    7,  -9,$41,$C0, 128
particle0B: .byte    0, -14,$50,$00      ; bubble small, palette 0,1,2,3
            .byte    5, -14,$51,$40
            .byte    0,  -9,$51,$80
            .byte    5,  -9,$51,$C0, 128
particle0C: .byte    0, -16,$40,$00      ; star bubble, palette 0,1,2,3
            .byte    7, -16,$41,$40
            .byte    0,  -9,$41,$80
            .byte    7,  -9,$41,$C0
            .byte    4, -13,$06,$02, 128

mascot:     .byte   120, 74,$60,$00      ; face mask for pico mascot
            .byte   132, 70,$61,$00
            .byte   125, 80,$62,$00, 128
valve:      .byte   160, 34,$63,$00      ; valve on pipe
            .byte   167, 34,$63,$40
            .byte   160, 41,$63,$80
            .byte   167, 41,$63,$C0, 128

particle_palette_table:
;      00   01   02   03   04   05   06   07   08   09   0A   0B   0C
.byte $23, $20, $20, $20, $21, $21, $20, $20, $23, $23, $23, $23, $23
PARTICLE_PALETTE_TABLE_SIZE = * - particle_palette_table

particle_sprite_table:
.word particle00, particle01, particle02, particle03
.word particle04, particle05, particle06, particle07
.word particle08, particle09, particle0A, particle0B
.word particle0C
PARTICLE_SPRITE_TABLE_SIZE = (* - particle_sprite_table) / 2

particle_table:
.byte $00, $00, $00, $01
.byte $02, $03, $04, $05
;.byte $06, $07, $08, $09
.byte $09, $0A, $0B, $0C

PARTICLE_TABLE_SIZE = * - particle_table
PARTICLE_TABLE_MASK = 1 << 4

.assert PARTICLE_PALETTE_TABLE_SIZE = PARTICLE_SPRITE_TABLE_SIZE, error, "particle table size mismatch"
.assert PARTICLE_TABLE_MASK >= PARTICLE_TABLE_SIZE, error, "particle table mask too small"

PARTICLE_COUNT = 16
.assert (PARTICLE_COUNT & (PARTICLE_COUNT-1)) = 0, error, "PARTICLE_COUNT must be power of 2"

.segment "ZEROPAGE"
temp_ptr: .res 2
oam_pos: .res 1

.segment "RAM"
particle_x: .res PARTICLE_COUNT ; X coordinate
particle_y: .res PARTICLE_COUNT ; Y coordinate
particle_p: .res PARTICLE_COUNT ; palette OR
particle_s: .res PARTICLE_COUNT ; speed (x/255 pixels per frame)
particle_t: .res PARTICLE_COUNT ; type
param_a: .res 1
param_b: .res 1
param_c: .res 1
speed_rng: .res 1
tick_cycle: .res 1
prng_seed: .res 2

.segment "CODE"

prng_reset:
	lda prng_seed+0
	ora prng_seed+1
	bne :+ ; already initialized
	lda nmi_count
	sta prng_seed+1
	ora prng_seed+0
	bne :+
		; make sure it's nonzero
		lda #$C5
		sta prng_seed+0
	:
	rts

prng:
	lda prng_seed+0
	asl
	rol prng_seed+1
	bcc :+
		eor #$BD
	:
	sta prng_seed+0
	rts

prng8:
	jsr prng
	jsr prng
	jsr prng
	jsr prng
	jsr prng
	jsr prng
	jsr prng
	jsr prng
	lda prng_seed+0
	rts

; X = particle
particle_spawn:
	; particle type
	jsr prng8
	:
		and #(PARTICLE_TABLE_MASK-1)
		cmp #PARTICLE_TABLE_SIZE
		bcc :+
		jsr prng
		jmp :-
	:
	tay
	lda particle_table, Y
	sta particle_t, X
	tay
	; X
	jsr prng8
	sta particle_x, X
	; Y
	lda #255
	sta particle_y, X
	; random palette selection
	jsr prng
	jsr prng
	jsr prng
	and particle_palette_table, Y
	sta particle_p, X
	; speed
	jsr prng8
	sta particle_s, X
	; if X is in middle, force attrib behind
	lda particle_x, X
	cmp #(115-23)
	bcc :+
	cmp #175
	bcs :+
		lda particle_p, X
		ora #$20
		sta particle_p, X
	:
	rts

wobble_06_07:
	lda particle_x, X
	clc
	adc nmi_count
	and #15
	bne :+
		lda particle_t, X
		eor #$01
		sta particle_t, X
	:
	lda particle_t, X
	rts

; X = particle
particle_tick:
	lda particle_s, X
	cmp speed_rng
	bcc :+
		dec particle_y, X
		beq particle_spawn
	:
	dec particle_y, X
	beq particle_spawn
	lda particle_t, X
	;cmp #$06
	;beq :+
	;cmp #$07
	;beq :+
	;jmp :++
	;:
	;	jsr wobble_06_07
	;:
	asl
	tay
	lda particle_sprite_table+0, Y
	sta temp_ptr+0
	lda particle_sprite_table+1, Y
	sta temp_ptr+1
	lda particle_x, X
	sta param_a
	lda particle_y, X
	sta param_b
	lda particle_p, X
	sta param_c
	txa
	pha
	jsr sprite_add
	pla
	tax
	rts

particles_reset:
	ldx #0
	:
		jsr particle_spawn
		jsr prng8
		sta particle_y, X
		inx
		cpx #PARTICLE_COUNT
		bcc :-
	; setup special palette
	lda #$0F
	sta palette+(4*4)+1 ; black for mascot eyes
	lda #$21
	sta palette+(6*4)+1 ; rainbow for star
	lda #$27
	sta palette+(7*4)+1 ; rainbow for star
	rts

rainbow_hue:
	lda palette, X
	pha
	and #$30
	sta palette, X ; preserve intensity
	pla
	and #$0F
	clc
	adc #1
	cmp #$0D
	bcc :+
		lda #$01
	:
	ora palette, X
	sta palette, X
	rts

tick_add: ; randomize tick order by incrementing by prime
.byte 1, 15, 3, 13, 7, 5, 11, 17

particles_tick:
	; cycle star rainbows
	lda nmi_count
	and #3
	bne @rainbow_end
		ldx #(6*4)+1
		jsr rainbow_hue
		ldx #(7*4)+1
		jsr rainbow_hue
	@rainbow_end:
	; bit reverse frame count for sub-pixel speed faking
	lda #0
	sta speed_rng
	lda nmi_count
	ldx #8
	:
		ror
		rol speed_rng
		dex
		bne :-
	; tick the particles in randomized order
	inc tick_cycle
	lda tick_cycle
	and #7
	sta tick_cycle
	lda #0
	:
		pha
		tax
		jsr particle_tick
		ldx tick_cycle
		pla
		clc
		adc tick_add, X
		and #(PARTICLE_COUNT-1)
		bne :-
	rts

sprite_begin:
	lda #$FF
	ldx #0
	:
		sta oam, X
		inx
		bne :-
	lda #4 ; skip sprite 0
	sta oam_pos
	rts

; param_a = X
; param_b = Y
; param_c = OR attribute
; temp_ptr = sprite
sprite_add:
	ldy #0
	ldx oam_pos
	bne :+
		rts ; OAM full
	:
		lda (temp_ptr), Y
		iny
		cmp #128
		beq :+
		clc
		adc param_a
		sta oam+3, X
		lda (temp_ptr), Y
		iny
		clc
		adc param_b
		sta oam+0, X
		lda (temp_ptr), Y
		iny
		sta oam+1, X
		lda (temp_ptr), Y
		ora param_c
		iny
		sta oam+2, X
		inx
		inx
		inx
		inx
		stx oam_pos
		beq :+ ; OAM full
		jmp :-
	:
	rts

first_sprites:
	lda #0
	sta param_a
	sta param_b
	sta param_c
	lda #<mascot
	sta temp_ptr+0
	lda #>mascot
	sta temp_ptr+1
	jsr sprite_add
	lda #0
	sta param_a
	sta param_b
	sta param_c
	lda #<valve
	sta temp_ptr+0
	lda #>valve
	sta temp_ptr+1
	jsr sprite_add
	rts

main:
	;
	; setup 
	;
	jsr prng_reset
	; palette
	ldx #0
	:
		lda fp_palette, X
		sta palette, X
		inx
		cpx #32
		bcc :-
	; nametable
	lda #<fp_title
	sta temp_ptr+0
	lda #>fp_title
	sta temp_ptr+1
	lda $2002
	lda #>$2000
	sta $2006
	lda #<$2000
	sta $2006
	ldy #0
	ldx #4
	:
		lda (temp_ptr), Y
		sta $2007
		iny
		bne :-
		inc temp_ptr+1
		dex
		bne :-
	; render particles
	jsr particles_reset
	jsr sprite_begin
	jsr first_sprites
	jsr particles_tick
	jsr ppu_update
	; main loop
@loop:
	; read gamepad
	jsr gamepad_poll
	; respond to gamepad state
	; TODO ?
@draw:
	; tick
	jsr sprite_begin
	jsr first_sprites
	jsr particles_tick
	jsr ppu_update
	; keep doing this forever!
	jmp @loop

;
; end of file
;
